/****************************************************************************
**
** Name: Xfer.c
**
** Description: This is a sample Active Message - LAPI C test program.
**
** Licensed Materials - Property of IBM
**
** 5765-296 5765-529 (C) Copyright IBM Corp. 1996, 1997
**
**  All rights reserved.
**  US Government Users Restricted Rights - Use, Duplication or Disclosure
**  restricted by GSA ADP Schedule Contract with IBM Corp.
**
*****************************************************************************/

/*
** Example Program illustrating use of the LAPI Active Message Call
*/

#include <pthread.h>
#include <lapi.h>

#define A_MAX     2
#define I_MAX     10

typedef struct    {              /* header for active message handler*/
    lapi_long_t   cmpl_hndlr;    /* pointer to completion handler */
    lapi_long_t   uinfo;         /* uinfo passed to  -- */
} usrhdr_t;                      /* the completion handler */

volatile int src_cont=0;
volatile int tgt_cont=0;

/* 
**
** hndl          Pointer to a LAPI context
** uhdr          Pointer to the user defined header to be handled
** uhdrlen       Pointer to the length of uhdr
** msglen        Pointer to the length of the message sent in the
**               call to LAPI_Xfer
** cmpl_hndlr    Pointer to the target side completion handler for
**               the message to be sent.
** saved_info    Pointer to any extra data in the user's header
**               that might be uselful in doing work in the
**               completion handler.
** hdr           Pointer to the user header. We need to use this
**               to satisfy the compiler. The compiler doesn't
**               know that uhdr is a pointer to a structure. It
**               thinks it's a pointer to a void.
*/
void *hdr_cmpl_hndlr(lapi_handle_t hndl, void *uhdr, uint *uhdrlen,
                     uint *msglen, compl_hndlr_t **cmpl_hndlr,
                     void** saved_info) {

   usrhdr_t   *hdr;
   /* Return to LAPI the pointer to the target side completion
      handler */
   hdr         = (usrhdr_t *)uhdr;
   *cmpl_hndlr = (compl_hndlr_t *)hdr->cmpl_hndlr;


   /* Return to LAPI a pointer to the the extra data sent in
      the user's header structure 
   *saved_info = (void *)&hdr->some_data;*/


   /* Update the global variable so that it can be checked in main,
      letting us know whether or not the header completion handler
      did indeed get invoked. 
   hdr_hndlr_cntr++;*/


   /* LAPI expects this header handler to return a pointer to the
      buffer where the data being sent is to be copied to. */
   return((void *)hdr->uinfo);
}

/*
** Function:     The send completion handler for the active call. This
**               is invoked at the src once the local buffer becomes 
**               available again.
** Parameters:   hndl ->  pointer to the handle for the LAPI context
**               param -> pointer to the user param (specified by user
**                        in the header handler function)
*/

void
send_compl_hndlr(lapi_handle_t *hndl, void *param)
{
    int loop, rc;
    int *buf;

    buf = (int *) param;
    printf("In Send Completion Handler, about to update cont\n");

    /* We set this to let the main routine know that the buffer
     * is available for reuse and that processing can continue.
     * (used instead of waiting on the origin counter)
     */
    src_cont = 1;
}




/*
** Function:     The completion handler for the active call. This
**               is invoked at the target after all the data of
**               the active message send (LAPI_Xfer) call have reached
**               the target.
** Parameters:   hndl ->  pointer to the handle for the LAPI context
**               param -> pointer to the user param (specified by user
**                        in the header handler function)
*/

void
tgt_compl_hndlr(lapi_handle_t *hndl, void *param)
{
    int loop, rc;
    int *buf;

    buf = (int *) param;
    printf("In Completion Handler: Result of Xfer call\n");

    /* Print Updated buffer */
    for (loop=0; loop < I_MAX; loop++) { 
        printf("val[%d] = %d\n", loop, buf[loop]);
    }

   /* Update counter so target may continue */
   tgt_cont = 1; 
}

/*
** Function:    User's active messsage header handler. This is invoked at the
**              target when the active message first arrives at the target
** Parameters:  hndl   -> pointer to the handle for the LAPI context
**              uhdr   -> pointer to the user header
**              uhdrlen -> pointer to the length of the user header
**              msglen -> pointer to the length of the message
**              compl_hndlr -> pointer to the completion handler function pointer
**                             This is be set by the user in this function
**                             (CAN be NULL)
**              saved_info -> pointer to the user_info. This is set by the user
**                            in this function. This paramter is then passed to
**                            the completion handler when the completion handler
**                            is invoked.
*/

void *
hdr_hndlr(lapi_handle_t *hndl, void *uhdr, uint *uhdrlen, uint *msglen,
          compl_hndlr_t **cmpl_hndlr, void **saved_info)
{
    void          *buf;
    usrhdr_t      *vhdr;

    printf("In Header Handler\n");
    vhdr             = (usrhdr_t *) uhdr;
    *cmpl_hndlr      = (compl_hndlr_t *) vhdr->cmpl_hndlr;
    *saved_info      = (void *) vhdr->uinfo;
    buf              = (void *) vhdr->uinfo;

    return (buf);
}


int
main(int argc, char **argv)
{
    lapi_handle_t t_hndl;               /* LAPI context handle - returned */
    lapi_info_t   t_info;               /* LAPI info structure */
    int           task_id,              /* My task id */
                  num_tasks;            /* Number of tasks in my job */
    lapi_cntr_t   l_cntr;               /* Origin counter */
    lapi_cntr_t   t_cntr;               /* Target counter */
    lapi_cntr_t   c_cntr;               /* Completion counter */
    int           t_buf[I_MAX];         /* Buffer to manipulate */
    lapi_long_t   global_addr[A_MAX];  /* Array to store t_buf addr from */
                                        /* all the tasks.  The size of this */
                                        /* array needs to each number of tasks */
    lapi_long_t   tgt_addr[A_MAX];     /* Array to store target counter addr */
                                        /* from all the tasks. */
    lapi_long_t   hndlr_addr[A_MAX];   /* Array to store header handlers */
    lapi_long_t   cmpl_hndlr_addr[A_MAX]; /* Address of completion handler */
    usrhdr_t      t_uhdr;               /* Store Header Handler information */
    void   *uhdr, *udata;
    int           uhdrlen, udatalen;
    int           loop, rc, tgt, val, cur_val;
    char          err_msg_buf[LAPI_MAX_ERR_STRING];
    lapi_xfer_t   t_xfer;                  /* store params to pass to Xfer routine */

    bzero(&t_info,sizeof(lapi_info_t));
    bzero(&t_xfer,sizeof(lapi_xfer_t));
    t_info.err_hndlr = NULL;   /* Not registering error handler function */
    
    if ((rc = LAPI_Init(&t_hndl, &t_info)) != LAPI_SUCCESS) {
        LAPI_Msg_string(rc, err_msg_buf);
        printf("Error Message: %s, rc = %d\n", err_msg_buf, rc);
        exit (rc);
    }
    
    rc = LAPI_Qenv(t_hndl, TASK_ID, &task_id);     /* Get task id within job */
    rc = LAPI_Qenv(t_hndl, NUM_TASKS, &num_tasks); /* Get no. of tasks in job */

    if (num_tasks != 2) {
        printf("Error Message: Program should run on 2 tasks\n");
        exit(1);
    }

    /* Turn off parameter checking - default is on */
    rc = LAPI_Senv(t_hndl, ERROR_CHK, 0); 

    /* Initialize counters to be zero at the start */
    rc = LAPI_Setcntr(t_hndl, &l_cntr, 0);
    rc = LAPI_Setcntr(t_hndl, &t_cntr, 0);
    rc = LAPI_Setcntr(t_hndl, &c_cntr, 0);

    /*
    ** Exchange buffer address, tgt_cntr address and hdr_hndlr address 
    ** and completion handler address of every task. Collective calls
    */
    rc = LAPI_Address_init64(t_hndl, (lapi_long_t) t_buf, global_addr);  
    rc = LAPI_Address_init64(t_hndl, (lapi_long_t)&t_cntr, tgt_addr);   
    rc = LAPI_Address_init64(t_hndl, (lapi_long_t)&hdr_hndlr, hndlr_addr);
    rc = LAPI_Address_init64(t_hndl, (lapi_long_t)&tgt_compl_hndlr, cmpl_hndlr_addr);


    if (task_id == 0) { /* Task id is 0 , Origin */
        tgt = task_id + 1;
        for (loop=0; loop < I_MAX; loop++) { /* Update buffer */
            t_buf[loop] = 2*loop;
        }
        rc = LAPI_Gfence(t_hndl);  /* Global fence to sync before starting */

        /* Fill in uhdr and udata buffers for AM call */
        t_uhdr.cmpl_hndlr  = (lapi_long_t)(cmpl_hndlr_addr[tgt]);
        t_uhdr.uinfo       = (lapi_long_t)(global_addr[tgt]);
        uhdr               = (void *)&t_uhdr;
        uhdrlen            = sizeof(usrhdr_t);
        udata              = (void *) t_buf;
        udatalen           = I_MAX*sizeof(int);
        
        t_xfer.Am.tgt = tgt;
        t_xfer.Am.hdr_hdl = (lapi_long_t) (hndlr_addr[tgt]);
        t_xfer.Am.uhdr = (void *) uhdr;
        t_xfer.Am.uhdr_len = uhdrlen;
        t_xfer.Am.udata = (void *) udata;
        t_xfer.Am.udata_len = udatalen;
        t_xfer.Am.shdlr  = (scompl_hndlr_t *) &send_compl_hndlr;
        t_xfer.Am.sinfo = (void *) 0;
        t_xfer.Am.org_cntr = &l_cntr;
        t_xfer.Am.cmpl_cntr = NULL;
        t_xfer.Am.tgt_cntr = (lapi_long_t) tgt_addr[tgt];
        t_xfer.Am.Xfer_type = LAPI_AM_XFER;
        rc = LAPI_Xfer(t_hndl, &t_xfer);

        /* Wait for local update from send completion handler*/
        while (src_cont == 0) {
            printf("waiting for send completion handler\n");
            sleep(1);
        }
    } else { /* Task id is 1 , Target */
        tgt = task_id - 1;
        for (loop=0; loop < I_MAX; loop++) { /* Zero out buffer */
            t_buf[loop] = 0;
        }
        rc = LAPI_Gfence(t_hndl);  /* Global fence to sync before starting */
        /* Process AM */
        rc = LAPI_Getcntr(t_hndl, &t_cntr, &val);
        while (val < 1) {
            sleep(1); /* Do some work */
            rc = LAPI_Probe(t_hndl); /* Poll the adapter once */
            rc = LAPI_Getcntr(t_hndl, &t_cntr, &val);
        }
        /* To reset the t_cntr value */
        /*rc = LAPI_Waitcntr(t_hndl, &t_cntr, 1, &cur_val);*/
        printf("Task %d, done doing work and processing AM\n", task_id);
        while (tgt_cont == 0) {
            sleep(1); /* Do some work */
        }
    }
    rc = LAPI_Gfence(t_hndl); /* Global fence to sync before terminating job */
    rc = LAPI_Term(t_hndl);
}



