

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/* 
** Example program to illustrate the use of the LAPI_Rmw64 api call.
** For a set of n tasks 0,1,...,n-1, where n is an even number, all
** tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
** Multiple calls are made to the Rmw64 routine, and the results of
** previous values are printed to show their effect on the target
** side.  A do_Rmw helper routine is provided for the repetetive
** code in the Rmw64 calls.  Note that this example is simply 
** the Rmw.c example modified to use LAPI_Rmw64.
** routine.
*/

/* 
** Used for the data in the Rmw calls. Since some calls use two
** values, we allocate space for two ints below.
*/
long long *data_buffer;

/* remote data buffer addrs */
void          **data_buffer_list;

/* Updates on src at message completion */
lapi_cntr_t   origin_cntr;

/* Constants to play with */
#define      SMALLERCONST   (long long) -128 
#define      BIGGERCONST    (long long)  256 

#ifndef      TRUE            
#define      TRUE                        1
#endif

/*
** Helper routine with repetetive code for Rmw call.  Args are those
** that are passed to Rmw (see descriptions below).
*/
int do_Rmw(lapi_handle_t handle, RMW_ops_t op, uint tgt, long long *tgt_var,
           long long *in_val, long long *prev_tgt_val) 
{
    lapi_cntr_t origin_cntr;  /* signifies message completion */
    int val;                  /* needed for waitcntr call */

    CHECK((LAPI_Setcntr(handle,&origin_cntr,0)));
    CHECK((LAPI_Rmw64(handle, op, tgt, tgt_var, in_val, prev_tgt_val,
                      &origin_cntr))); 
    CHECK((LAPI_Waitcntr(handle, &origin_cntr,1, &val)));

    return LAPI_SUCCESS;
}


int main( int argc, char **argv )
{

lapi_handle_t handle;           /* LAPI handle */
lapi_info_t   info;             /* Info to pass to LAPI_Init */
int           task_id;          /* Our LAPI Task id */
int           num_tasks;        /* Total number of tasks */
int           i,j;              /* Loop counters */
int           buddy;            /* Our communication partner */
long long     prev_val;         /* Used in Rmw call */

    /* Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(-1);
    }

    /* target uses two var's for COMPARE_AND_SWAP. */
    data_buffer = (long long *) malloc(2*sizeof(long long));

    /* allocate buffer address table */
    data_buffer_list = (void **) malloc(num_tasks*sizeof(void *));

    /*
    ** collective calls.  each task stores its own addr plus remote
    ** addrs
    */
    CHECK((LAPI_Address_init(handle,(void *)data_buffer,data_buffer_list)));

    /*
    ** up to this point, all instructions have executed on all tasks.
    ** we now begin differentiating tasks.
    */
    if ( task_id % 2  == 0 ){   /* sender */

        /* setup buddy pairs as 0-1, 2-3, 4-5, etc.*/
        buddy = task_id + 1;

        /* set up initial values to play with */
        data_buffer[0] = SMALLERCONST;
        data_buffer[1] = (long long) 0;

        /* collective call.  sync before starting data transfer */
        CHECK((LAPI_Gfence(handle)));

        /*
        ** Execute a FETCH_AND_ADD Rmw64 call.  Buddy inits his values to
        ** 0, so a FETCH_AND_ADD of our value should cause his val to
        ** match ours (currently SMALLERCONST).  After the call, we
        ** verify that his buffer was initially 0 by printing the
        ** value passed back to us in prev_val.  Note that our do_Rmw
        ** helper handles all the counter logic.  See that routine for
        ** the actual arglist to the LAPI_Rmw64 call.
        ** args:  handle                   -- the LAPI handle
        **        FETCH_AND_ADD            -- Rmw operation for this call
        **        buddy                    -- the target task id
        **        data_buffer_list[buddy]  -- addr of remote var
        **        &data_buffer[0]          -- addr of local var
        **        &prev_val                -- ref var to return previous
        **                                    data value on target
        */
        CHECK((do_Rmw(handle, FETCH_AND_ADD, buddy,
                          data_buffer_list[buddy], &data_buffer[0],
                          &prev_val)));  

        /* Show buddy's previous result */
        printf("FETCH_AND_ADD prev_val: %lld, expected: %lld\n",prev_val,(long long)0);

        /* 
        ** Execute a COMPARE_AND_SWAP call.  Since local buffer and
        ** remote buffer values are equal, prev_tgt should be TRUE
        ** and tgt buffer should get value of data_buffer[1]; we can
        ** confirm TRUE value now and new value of prev_val in next call.
        */
        data_buffer[1] = BIGGERCONST;
        CHECK((do_Rmw(handle, COMPARE_AND_SWAP, buddy,
                           data_buffer_list[buddy], &data_buffer[0], 
                            &prev_val)));  

        /* Show buddy's previous result */
        printf("COMPARE_AND_SWAP prev_val: %lld, expected: %lld\n",
                prev_val,(long long)TRUE);

        /* 
        ** Execute a SWAP call.  Buddy should get SMALLERCONST.
        */
        CHECK((do_Rmw(handle, SWAP, buddy,
                           data_buffer_list[buddy], &data_buffer[0],
                           &prev_val)));

        /* confirm results of first SWAP */ 
        printf( "First SWAP prev_val: %lld, expected: %lld\n",
                prev_val,BIGGERCONST);
             
        /* swap once more just to confirm value on remote side */
        CHECK((do_Rmw(handle, SWAP, buddy, 
                           data_buffer_list[buddy], &data_buffer[0],
                           &prev_val)));

        /* confirm results of second SWAP */ 
        printf( "Second SWAP prev_val: %lld, expected: %lld\n",
                prev_val,SMALLERCONST);

    } else {             /* receiver */
        buddy = task_id - 1;

        /* Initialize data buffer */
        data_buffer[0] = (long long)0;
        data_buffer[1] = (long long)0;

        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));

    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(data_buffer);
    free(data_buffer_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


