

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

#define VECLEN         15
#define NUM_VECS       10
#define ARRAYLEN       10

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];



/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Accumulate_and_return Example Program using the 
**   LAPI_Amsend api call.  For a set of n tasks 0,1,...,n-1,
**   where n is an even number, all tasks are divided into
**   (src,tgt) buddy pairs (0,1), (2,3), etc.  For each pair,
**   an array of ints is transferred from src to tgt with a
**   LAPI_Amsend call.  The ints are then manipulated
**   and sent back to the src with a LAPI_Amsend call from
**   within the completion handler.  Synchronization in this
**   example is driven by the fact that the target makes an
**   Amsend call back to the origin from within the target's
**   completion handler.  The origin can wait on a simple flag
**   that gets set in its completion handler (called when Am
**   from target completes).  The target passes a completion
**   counter to its Am call then waits for the counter to
**   increment.  The various data buffers are printed along
**   the way to show progress.
*/


/* Data structure for user header */
typedef struct {
    uint length;
    uint buddy;
} uhdr_t;

/* Data structure for info to pass to completion handler */
typedef struct {
    int *tmp_buf;
    uint length;
    uint buddy;
} compl_info_t;

compl_info_t compl_info;

/* origin vector on src, target vector on tgt */
lapi_vec_t *vec;

/* stores initial value on src and tgt and final value on tgt */
int           data_buffer[ARRAYLEN];

/* global array of header handler addrs */
void   **hdr_hdl_list;


/*
** used on tgt. waits for return of completion handler
** on src after second Am call 
*/
lapi_cntr_t   compl_cntr;


/* counter used on src (not a LAPI cntr) */
int wait_flag;


/*
**  src side handlers (invoked when tgt does return
**  Amsend call.
*/

/* completion handler on original src (as final tgt) */
void set_final_cntr(lapi_handle_t *handle, void *param)
{
    int i;     /* loop counter */

    printf("Greetings from the src completion handler\n");
    for( i=0; i<compl_info.length; i++ ) {
        printf("final data_buf[%d]: %d\n", i, data_buffer[i]);
    }
  
    /* set wait flag here to continue execution below */
    wait_flag++;

}

/* header_handler on src */
void *src_header_handler(lapi_handle_t *handle, void *uhdr,
                         uint uhdr_len, uint msg_len,
                         compl_hndlr_t **completion_handler,
                         void **user_info)
{
    uhdr_t *uhdr_ptr = (uhdr_t *) uhdr;  /* info from src */

    printf("Greetings from the src header handler...\n");

    *completion_handler = set_final_cntr;
    compl_info.length = uhdr_ptr->length;

    /* LAPI will write data payload here */
    return data_buffer;
}

/*
**  tgt side handlers (invoked when src does return
**  Amsend call.
*/

/* completion handler on original target */
void do_accumulate(lapi_handle_t *handle, void *param)
{
    uhdr_t        uhdr;      /* pass info to HH on src */
    int i, j;                /* loop counters */
    int buddy;               /* we send to orig. src */
    uint len;                /* pulled from param */
    int *tmp_buf;            /* pulled from param */

    /* grab info from compl_info */
    tmp_buf = (int *)(compl_info.tmp_buf);
    len     = (uint) (compl_info.length);
    buddy   = (uint) (compl_info.buddy);

    printf("Greetings from the tgt completion handler...\n");

    /* do the accumulation */
    for( i = 0; i < len; i++ ) {
        data_buffer[i] += tmp_buf[i];
    }

    /* setup for the return Amsend */
    uhdr.length = len;

    CHECK((LAPI_Amsend(*handle, buddy, 
                       (void *)(hdr_hdl_list[buddy]), &uhdr,
                        sizeof(uhdr_t), &(data_buffer[0]),
                        len*(sizeof(data_buffer[0])),
                        NULL, NULL, &compl_cntr)));

    /* done with tmp_buf */
    free(tmp_buf);

    free(vec->info);
    free(vec->len);
    free(vec);
}

/* header handler on original tgt */
/* header handler that runs on target when first packet arrives.
**   sets the completion_handler pointer and returns the address
**   of the target vector.
*/
lapi_vec_t *tgt_header_handler(lapi_handle_t *handle, void *uhdr,
                           uint uhdr_len, ulong *len_vec[],
                           compl_hndlr_t **completion_handler,
                           void **user_info)
{
    int i;                   /* loop counter */
    uhdr_t *uhdr_ptr;        /* for info from uhdr */
    uint len;                /* pulled from user header */
    uint buddy;              /* pulled from uhdr */ 
    int *tmp_buf;            /* intermediate buffer for accumulate */



    printf("Greetings from the tgt header handler...\n");

    /* initialize the vector whose address will be returned to LAPI */
    vec = (lapi_vec_t *) malloc(sizeof(lapi_vec_t));

    /*
    ** set the fields to match those of the origin vector. allocate
    ** storage for the length and data arrays.  malloc'ed data will be
    ** freed in the completion handler.
    */
    vec->num_vecs = NUM_VECS;
    vec->vec_type = LAPI_GEN_IOVECTOR;
    vec->len = (unsigned long *)malloc(NUM_VECS*sizeof(unsigned long));
    vec->info = (void **) malloc(NUM_VECS*sizeof(void *));
    for( i=0; i < NUM_VECS; i++ ) {
        vec->info[i] = (void *) &data_buffer[i];
        vec->len[i] = (unsigned long)(sizeof(int));
    }


    /* grab the info passed in the user header */
    uhdr_ptr = (uhdr_t *) uhdr;
    len = uhdr_ptr->length;
    buddy = uhdr_ptr->buddy;

    /*
    ** malloc to buffer then pass back to LAPI for use in
    ** completion handler.  completion handler will free.
    */
    tmp_buf = (int *) malloc(len*sizeof(int));

    /* store some arbitrary values in tmp_buf */
    for (i = 0; i < len; i++ ) {
        tmp_buf[i] = (len) - (2*i);
    }

    /* set completion handler pointer */
    *completion_handler = do_accumulate;

    /* set params to send to completion handler */
    compl_info.tmp_buf = tmp_buf;
    compl_info.length = len;
    compl_info.buddy = buddy;

    return vec;
    
}

int main( int argc, char **argv )
{

lapi_handle_t handle;                      /* LAPI handle */
lapi_info_t   info;                        /* Info to pass to
                                              LAPI_Init */
int           task_id;                     /* Our LAPI Task id */
int           num_tasks;                   /* Total number of tasks */
void          *buf_addr_list[NUM_TASKS];   /* Table of remote data
                                              buf addrs. */
int           i;                           /* Loop counter */
int           val;                         /* Needed for waitcntr call */
int           buddy;                       /* Our communication
                                              partner */

    /*
    ** Clear the struct.  Not passing any info to init through this
    ** struct.
    */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks % 2) != 0 ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    /* setup for header handler addresses */
    hdr_hdl_list = (void **) malloc(num_tasks*sizeof(void *));

    /*
    ** up to this point, all instructions have executed on all 
    ** tasks. we now begin differentiating tasks.
    */
    if ((task_id %2) == 0) {   /* message source */
        uhdr_t        uhdr;

    /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id+1;

    /*
    ** collective calls.  in this case, we need to specify
    ** different header handlers for the src and tgt tasks,
    ** thus the Address_init call is done within the separate 
    ** code for each task.
    */
        CHECK((LAPI_Address_init(handle,(void *)&src_header_handler,hdr_hdl_list)));

    /* for our own sync. */
        wait_flag = 0;

    /* initialize data buffer */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i;
            printf("original data_buffer[%d]: %d\n", i, data_buffer[i]);
        }
 
   /* initialize the vector whose address will be returned to LAPI */
        vec = (lapi_vec_t *) malloc(sizeof(lapi_vec_t));

    /*
    ** set the fields to match those of the origin vector. allocate
    ** storage for the length and data arrays.  malloc'ed data will be
    ** freed in the completion handler.
    */
        vec->num_vecs = NUM_VECS;
        vec->vec_type = LAPI_GEN_IOVECTOR;
        vec->len = (unsigned long *)
                    malloc(NUM_VECS*sizeof(unsigned long));
        vec->info = (void **) malloc(NUM_VECS*sizeof(void *));
        for( i=0; i < NUM_VECS; i++ ) {
            vec->info[i] = (void *) &data_buffer[i];
            vec->len[i] = (unsigned long)(sizeof(int));
        }

        /* sync before starting */
        CHECK((LAPI_Gfence(handle)));

        /* initialize our header */
        uhdr.length = ARRAYLEN;

        /* tgt will use in return Amsend call. */
        uhdr.buddy = task_id;  


    /*
    ** execute the data transfer to our buddy task.  send ARRAYLEN ints,
    ** starting with data_buffer[0].  Note that no user header is sent
    ** and no counters are used.
    ** args:  handle                 -- the LAPI handle
    **        buddy                  -- the target task id
    **        hdr_hndlr_list[buddy]  -- target header handler
    **        &uhdr                  -- the user header
    **        sizeof(uhdr_t)         -- the user header length
    **        vec                    -- the origin vector
    **        NULL                   -- target counter
    **        NULL                   -- origin counter
    **        NULL                   -- completion counter
    */
        CHECK((LAPI_Amsendv(handle, buddy, (void *)
                                        hdr_hdl_list[buddy], &uhdr,
                                        sizeof(uhdr_t), vec, NULL,
                                        NULL, NULL)));


        /* wait for src compl hndlr to signal that we're done */
        while (!wait_flag) {
            sleep(1);
        }
        


    } else {             /* receiver */

        /*
        ** collective calls.  in this case, we need to specify
        ** different header handlers for the src and tgt tasks,
        ** thus the Address_init call is done within the separate
        ** code for each task.
        */
        CHECK((LAPI_Address_init(handle,(void *)&tgt_header_handler,
                                             hdr_hdl_list)));

        /*
        ** initialize the completion counter. When it increments, we know
        ** that the completion handler on the src has returned.  Our
        ** completion handler (target side) will pass this counter to
        ** its Am call.
        */
        CHECK((LAPI_Setcntr(handle,&compl_cntr,0)));

        /* sync with src */
        CHECK((LAPI_Gfence(handle)));

        /* wait for completion handler on origin to complete */
        CHECK((LAPI_Waitcntr(handle, &compl_cntr,1, &val)));
    }

    /* remaining code will execute on all tasks */ 

    /* final sync before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(hdr_hdl_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


